# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/02_github.ipynb.

# %% auto 0
__all__ = ['bcolors', 'GitHubGroup']

# %% ../nbs/api/02_github.ipynb 3
from github import Github
import github
import json
import time
import os
import glob
from pprint import pprint

# %% ../nbs/api/02_github.ipynb 4
class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

# %% ../nbs/api/02_github.ipynb 5
class GitHubGroup:
    def __init__(self,
                 credentials_fp="", # the file path to the credential json
                 org="", # the organization name
                 verbosity=1 # Controls the verbosity: 0=silent, 1=print status
                ):
        self.github = None
        self.org = None
        self.verbosity = verbosity
        
        if credentials_fp != "":
            self.auth_github(credentials_fp)
        if org != "":
            self.set_org(org)

    def auth_github(self,
                    credentials_fp: str # the personal access token generated at GitHub Settings
                   ):
        "Authenticate GitHub account with the provided credentials"
        with open(credentials_fp, "r") as f:
            token = json.load(f)["GitHub Token"]
        self.github = Github(token)
        # check authorization
        _ = self.github.get_user().get_repos()[0]
        if self.verbosity != 0:
            print(f"Successfully Authenticated. "
                  f"GitHub account: {bcolors.OKGREEN} {self.github.get_user().login} {bcolors.ENDC}")
        
    def set_org(self,
                org: str # the target organization name
               ):
        "Set the target organization for repo creation"
        self.org = self.github.get_organization(org)
        if self.verbosity != 0:
            print(f"Target Organization Set: {bcolors.OKGREEN} {self.org.login} {bcolors.ENDC}")

    def create_repo(self,
                    repo_name: str, # repository name
                    repo_template="", # template repository that new repo will use. If empty string, an empty repo will be created. Put in the format of "<owner>/<repo>"
                    private=True, # visibility of the created repository
                    description="", # description for the GitHub repository
                    personal_account=False, # create repos in personal GitHub account
                    ) -> github.Repository.Repository:
        "Create a repository, either blank, or from a template"
        if self.org is None and personal_account:
            raise ValueError("Organization is not set")
        if personal_account:
            parent = self.github.get_user()
        else:
            parent = self.org
        if repo_template == "":
            return parent.create_repo(
                name=repo_name,
                private=private,
                description=description
            )
        # create from template
        return parent.create_repo_from_template(
            name=repo_name,
            repo=self.get_repo(repo_template),
            private=private,
            description=description,
        )
    
    def get_repo(self,
                 repo_full_name: str # full name of the target repository
                ) -> github.Repository.Repository:
        "To get a repository by its name"
        try:
            return self.github.get_repo(repo_full_name)
        except Exception:
            return self.org.get_repo(repo_full_name)
    
    def get_org_repo(self,
                     repo_full_name: str # full name of the target repository
                    ) -> github.Repository.Repository:
        "Get a repository within the target organization"
        return self.org.get_repo(repo_full_name)

    
    def get_team(self,
                 team_slug:str # team slug of the team
                ) -> github.Team.Team:
        "Get the team inside the target organization"
        if self.org is None:
            raise ValueError("The organization has not been set. Please set it via g.set_org")
        return self.org.get_team_by_slug(team_slug)
    
    def rename_files(self,
                     repo: github.Repository.Repository, # the repository that we want to rename file
                     og_filename: str, # old file name
                     new_filename: str # new file name
                    ):
        "Rename the file by delete the old file and commit the new file"
        file = repo.get_contents(og_filename)
        repo.create_file(new_filename, "rename files", file.decoded_content)
        repo.delete_file(og_filename, "delete old files", file.sha)
        if self.verbosity != 0:
            print(f"File Successfully Renamed from "
                  f" {bcolors.OKCYAN} {og_filename} {bcolors.ENDC} "
                  f" to {bcolors.OKGREEN} {new_filename} {bcolors.ENDC}")
        
    def add_collaborator(self,
                          repo: github.Repository.Repository, # target repository
                          collaborator:str, # GitHub username of the collaborator
                          permission:str # `pull`, `push` or `admin`
                         ):
        "Add collaborator to the repository with specified permission"
        try:
            repo.add_to_collaborators(collaborator, permission)
        except Exception as e:
            print(f"{bcolors.WARNING}Add Failed for {collaborator}{bcolors.ENDC}")
        if self.verbosity != 0:
            print(f"Added Collaborator: {bcolors.OKGREEN} {collaborator} {bcolors.ENDC}"
                  f" to: {bcolors.OKGREEN} {repo.name} {bcolors.ENDC} with "
                  f"permission: {bcolors.OKGREEN} {permission} {bcolors.ENDC}")
    
    def remove_collaborator(self,
                            repo: github.Repository.Repository, # target repository
                            collaborator:str, # GitHub username of the collaborator
                           ):
        "Remove collaborator privileges from the repository"
        repo.remove_from_collaborators(collaborator)
        
    def resend_invitations(self,
                          repo: github.Repository.Repository, # target repository
                         ) -> [github.NamedUser.NamedUser]: # list of re-invited user
        "Resent Invitation to invitee who did not accept the invitation"
        pendings = list(repo.get_pending_invitations())
        users = [p.invitee for p in pendings]
        if self.verbosity != 0:
            print("The list of pending invitation:")
            pprint(users)
        for p in pendings:
            repo.remove_invitation(p.id)
            if self.verbosity != 0:
                print(f"{bcolors.WARNING}{bcolors.UNDERLINE}{p.invitee.login}{bcolors.ENDC} {bcolors.FAIL}Invite Revoked {bcolors.ENDC}")
            self.add_collaborator(repo, p.invitee.login, p.permissions)
            if self.verbosity != 0:
                print(f"{bcolors.OKGREEN} Invite Resent to {p.invitee.login} {bcolors.ENDC}")
        return users

    def resent_invitations_team_repos(self,
                                      team_slug: str # team slug (name) under the org
                                     ):
        "For all repository under that team, Resent invitation to invitee who did not accept the invitation"
        team = self.get_team(team_slug)
        repos = team.get_repos()
        for repo in repos:
            if self.verbosity != 0:
                print(f"Repository {bcolors.OKCYAN} {repo.name} {bcolors.ENDC}:")
            try:
                _ = self.resend_invitations(repo)
            except Exception as e:
                print(f"{bcolors.WARNING}Make sure to have proper rights to the target repo{bcolors.ENDC}\n")
                print(e)
        
    def add_team(self,
                  repo: github.Repository.Repository, # target repository
                  team_slug: str, # team slug (name)
                  permission:str # `pull`, `push` or `admin`
                 ):
        "Add team to the repository with specified permission"
        team = self.get_team(team_slug)
        team.add_to_repos(repo)
        team.update_team_repository(repo, permission)
        if self.verbosity != 0:
            print(f"Team {bcolors.OKGREEN} {team.name} {bcolors.ENDC} "
                  f"added to {bcolors.OKGREEN} {repo.name} {bcolors.ENDC} "
                  f"with permission {bcolors.OKGREEN} {permission} {bcolors.ENDC}")

    def create_feedback_dir(self,
                            repo: github.Repository.Repository, # target repository
                            template_fp: str,
                            destination="feedback" # directory path of the template file.
                           ):
        "Create feedback directory on local machine"
        os.makedirs(destination, exist_ok=True)
        os.makedirs(f"{destination}/{repo.name}", exist_ok=True)
        files = glob.glob(f"{template_fp}/*")
        for file in files:
            head = os.path.split(file)[1]
            with open(file, "r") as f:
                file = f.read()
            with open(f"{destination}/{repo.name}/{head}", "w+") as f:
                f.write(file)
            if self.verbosity != 0:
                print(f"File {bcolors.OKGREEN}{head}{bcolors.ENDC} "
                      f"created at {bcolors.OKGREEN}{destination}/{repo.name}{bcolors.ENDC}"
                     )

    def create_issue(self,
                     repo: github.Repository.Repository, # target repository
                     title: str, # title of the issue,
                     content: str # content of the issue
                    ) -> github.Issue.Issue: # open issue
        "Create GitHub issue to the target repository"
        issue = repo.create_issue(title=title, body=content)
        if self.verbosity != 0:
            print(f"In the repo: {bcolors.OKGREEN}{repo.name}{bcolors.ENDC},")
            print(f"Issue {bcolors.OKGREEN}{title}{bcolors.ENDC} Created!")
        return issue
    
    def create_issue_from_md(self,
                             repo: github.Repository.Repository, # target repository,
                             md_fp: str # file path of the feedback markdown file
                            ) -> github.Issue.Issue: # open issue
        "Create GitHub issue from markdown file."
        md = ""
        with open(md_fp, "r") as f:
            md = f.read()
        title = md.split("\n")[0][2:]
        content = md
        return self.create_issue(repo, title, content)
    
    def release_feedback(self,
                         md_filename: str, # feedback markdown file name
                         feedback_dir="feedback", # feedback directory contains the markdown files
                        ):
        "Release feedback via GitHub issue from all the feedbacks in the feedback directory"
        repo_names = os.listdir(feedback_dir)
        for repo_name in repo_names:
            if repo_name == ".DS_Store":
                continue
            try:
                repo = self.org.get_repo(repo_name)
            except Exception:
                print(f"Repo: {bcolors.WARNING}{repo_name} NOT FOUND!{bcolors.ENDC}")
            self.create_issue_from_md(repo, os.path.join(feedback_dir, repo_name, md_filename))
        
    def create_group_repo(self,
                          repo_name: str, # group repository name
                          collaborators: [str], # list of collaborators GitHub id
                          permission: str, # the permission of collaborators. `pull`, `push` or `admin`
                          rename_files=dict(), # dictionary of files renames {<og_name>:<new_name>}
                          repo_template="", # If empty string, an empty repo will be created. Put in the format of "<owner>/<repo>"
                          private=True, # visibility of the created repository
                          description="", # description for the GitHub repository
                          team_slug="", # team slug, add to this repo
                          team_permission="", # team permission to this repository `pull`, `push` or `admin`
                          feedback_dir=False, # whether to create a feedback directory for each repository created
                          feedback_template_fp="", # the directory of the feedback template
                         ) -> github.Repository.Repository: # created repository
        "Create a Group Repository"
        repo = self.create_repo(repo_name, repo_template, private, description)
        if self.verbosity != 0:
            print(f"Repo {bcolors.OKGREEN} {repo.name} {bcolors.ENDC} Created... Wait for 3 sec to updates")
        time.sleep(3)
        for og_name, new_name in rename_files.items():
            self.rename_files(repo, og_name, new_name)
        for collaborator in collaborators:
            self.add_collaborator(repo, collaborator, permission)
        if team_slug != "":
            self.add_team(repo, team_slug, team_permission)
        if self.verbosity != 0:
            print(f"Group Repo: {bcolors.OKGREEN} {repo_name} {bcolors.ENDC} successfuly created!")
            print(f"Repo URL: https://github.com/{self.org.login}/{repo_name}")
        if feedback_dir:
            if feedback_template_fp == "":
                raise ValueError("You have to specify the template files.")
            self.create_feedback_dir(repo, template_fp=feedback_template_fp)
        return repo




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/04_project_grading.ipynb.

# %% auto 0
__all__ = ['bcolors', 'Grading']

# %% ../nbs/api/04_project_grading.ipynb 3
from .github import GitHubGroup
from .canvas import CanvasGroup
import github
import canvasapi
from ast import literal_eval

# %% ../nbs/api/04_project_grading.ipynb 4
class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

# %% ../nbs/api/04_project_grading.ipynb 5
class Grading:
    def __init__(self,
                 ghg:GitHubGroup=None, # authenticated GitHub object
                 cg:CanvasGroup=None, # authenticated canvas object
                ):
        self.ghg = ghg
        self.cg = cg

    def create_issue_from_md(self,
                              repo:github.Repository.Repository, # target repository to create issue
                              md_fp: str # file path of the feedback markdown file
                              ) -> github.Issue.Issue: # open issue
        "Create GitHub issue from markdown file."
        return self.ghg.create_issue_from_md(repo, md_fp)

    def fetch_issue(self,
                    repo:github.Repository.Repository, # target repository to fetch issue
                    component:str, # the component of the project grading, let it be proposal/checkpoint/final. Need to match the issue's title
                    ) -> github.Issue.Issue:
        "Fetch the issue on GitHub repo and choose related one"
        for issue in list(repo.get_issues()):
            if component.lower() in issue.title.lower():
                return issue
        raise ValueError(f"Issue related to {component} did not found.")

    def parse_score_from_issue(self,
                               repo:github.Repository.Repository, # target repository to create issue
                               component:str, # The component of the project grading, let it be proposal/checkpoint/final. Need to match the issue's title
                               ) -> int: # the fetched score of that component
        "parse score from the template issue"
        issue = self.fetch_issue(repo, component)
        body = issue.body
        score = 0
        for line in body.split("\n"):
            if "Score =" in line and "[comment]" not in line:
                score = literal_eval(line.split("=")[1])
                return score
        raise ValueError(f"Score Parse Error. please check the score format on github. \n"
                         f"Issue URL: {issue.url}")

    def update_canvas_score(self,
                            group_name:str, # target group name on a canvas group
                            assignment_id, # assignment id of the related component
                            score:float, # score of that component
                            issue:github.Issue.Issue=None,
                            post=False, # whether to post score via api. for testing purposes
                            ):
        "Post score to canvas"
        if self.cg.group_category is None:
            raise ValueError("CanvasGroup's group_category not set.")
        members = self.cg.group_to_emails[group_name]
        self.cg.link_assignment(assignment_id)
        for member in members:
            student_id = self.cg.email_to_canvas_id[member]
            text_comment = f"Group: {group_name}"
            if issue is not None:
                text_comment += f"\nView at {issue.url.replace('https://api.github.com/repos', 'https://github.com')}"
            if post:
                self.cg.post_grade(
                    student_id=student_id,
                    grade=score,
                    text_comment=text_comment
                )
            else:
                print(f"{bcolors.WARNING}Post Disable{bcolors.ENDC}")
                print(f"For student: {member}, the score is {score}")
                print(f"Comments: {text_comment}")

    def grade_project(self,
                      repo:github.Repository.Repository, # target repository to grade
                      component:str, # The component of the project grading, let it be proposal/checkpoint/final. Need to match the issue's title
                      assignment_id:int, # assignment id that link to that component of the project
                      canvas_group_name:dict=None, # mapping from GitHub repo name to Group name. If not specified, the repository name will be used.
                      canvas_group_category:str=None, # canvas group category (set)
                      post:bool=False, # whether to post score via api. For testing purposes
                      ):
        "grade github project components"
        # set the category if you haven't
        if canvas_group_category is not None:
            self.cg.set_group_category(canvas_group_category)
        score = self.parse_score_from_issue(repo, component)
        # create mapping from GitHub repo name to canvas group name
        if canvas_group_name is not None:
            group_name = canvas_group_name[repo.name]
        else:
            group_name = repo.name
        if score is ...:
            print(f"{bcolors.WARNING}{group_name}'s {component} Not Graded. {bcolors.ENDC}")
            return
        issue = self.fetch_issue(repo, component)
        self.update_canvas_score(group_name, assignment_id, score, issue, post)





# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/01_GroupEng_assign.ipynb.

# %% auto 0
__all__ = ['AssignGroup']

# %% ../nbs/api/01_GroupEng_assign.ipynb 3
import GroupEng
import canvasapi
import github
from . import GitHubGroup, CanvasGroup

# %% ../nbs/api/01_GroupEng_assign.ipynb 4
class AssignGroup:
    def __init__(self,
                 ghg: GitHubGroup, # authenticated GitHub object
                 cg: CanvasGroup, # authenticated canvas object
                 groupeng_config="", # Directory for the GroupEng config yml file
                ):
        "Initializer for Assign Group"
        self.status = None
        self.out_dir = None
        self.prefix = None
        self.cg = cg
        self.ghg = ghg
        # Initialize if appropriate parameters are defined
        if groupeng_config != "":
            self.assign_groups(groupeng_config)

    def assign_groups(self,
                      groupeng_config:str, # Directory for the GroupEng config yml file
                      assign_canvas_group=False, # directly assign canvas groups
                      create_gh_repo=False, # directly create GitHub repos
                      username_quiz_id=-1, # username quiz id from canvas course
                      in_group_category="", # specify which group category the group belongs to
                      suffix="", # suffix to the group name
                     ) -> (bool, str): # Status and output directory of the compiled file.
        status, out_dir = GroupEng.run(groupeng_config)
        self.status, self.out_dir = status, out_dir
        file = os.path.split(groupeng_config)[1]
        self.prefix = os.path.splitext(file)[0]
        if assign_canvas_group:
            if self.cg.group_category is None and in_group_category == "":
                raise ValueError("Have to specify in_group_category to create canvas group")
            self.create_canvas_group(in_group_category, suffix)
        if create_gh_repo:
            if username_quiz_id == -1:
                raise ValueError("Have to specify the canvas username quiz id")
            self.create_github_group(username_quiz_id)
        return status, out_dir

    def create_canvas_group(self,
                            in_group_category="", # specify which group category the group belongs to
                            suffix="", # suffix to the group name
                            ):
        "Create canvas groups based on the generated group configuration"
        if self.out_dir is None:
            raise ValueError("The group configuration has not been set. Please assign group via assign_groups")
        if self.cg.group_category is None:
            raise ValueError("The group category has not been set.")
        if in_group_category == "":
            in_group_category = self.cg.group_category.name
        # load the generated configuration file
        groups_generated_fp = os.path.join(self.out_dir, f"{self.prefix}_groups.csv")
        with open(groups_generated_fp, "r") as f:
            groups = f.read().splitlines()
        # create canvas groups for each.
        for group in groups:
            group = group.replace(" ", "").split(",")
            group_name, group_members = group[0], group[1:]
            self.cg.assign_canvas_group(
                group_name=f"{group_name}{suffix}",
                group_members=group_members,
                in_group_category=in_group_category
            )

    def create_github_group(self,
                            username_quiz_id:int # username quiz id from canvas course
                            ):
        github_usernames = self.cg.fetch_username_from_quiz(username_quiz_id)
        self.cg.set_group_category(cg.group_category.name)
        groups = self.cg.group_to_emails
        repos = []
        for group_name, members in groups.items():
            group_git_usernames = []
            for email in members:
                try:
                    # try to get the git username for each student.
                    # not all students completed their quiz.
                    group_git_usernames.append(github_usernames[email])
                except KeyError:
                    print(f"{email}'s GitHub Username not found")
            repo = self.ghg.create_group_repo(
                repo_name=group_name,
                collaborators=group_git_usernames,
                permission="write",
                repo_template="COGS118A/group_template",
                rename_files={
                    "Checkpoint_groupXXX.ipynb": f"Checkpoint_{group_name}.ipynb",
                    "FinalProject_groupXXX.ipynb": f"FinalProject_{group_name}.ipynb",
                    "Proposal_groupXXX.ipynb": f"Proposal_{group_name}.ipynb"
                },
                private=True,
                description=f"COGS118A Final Project {group_name} Repository",
                team_slug="Instructors_Sp23",
                team_permission="admin"
            )
            print("")
            repos.append(repo)
        return repos





# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/api/03_canvas.ipynb.

# %% auto 0
__all__ = ['bcolors', 'CanvasGroup']

# %% ../nbs/api/03_canvas.ipynb 3
from canvasapi import Canvas
from github import Github
import canvasapi
import json
import requests
import time
import sys
import numpy as np
import pandas as pd
from io import StringIO

# %% ../nbs/api/03_canvas.ipynb 4
class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

# %% ../nbs/api/03_canvas.ipynb 5
class CanvasGroup():
    def __init__(self,
                 credentials_fp = "", # credential file path. [Template of the credentials.json](https://github.com/FleischerResearchLab/CanvasGroupy/blob/main/nbs/credentials.json)
                 API_URL="https://canvas.ucsd.edu", # the domain name of canvas
                 course_id="", # Course ID, can be found in the course url
                 group_category="", # target group category (set) of interests
                 verbosity=1 # Controls the verbosity: 0 = Silent, 1 = print all messages
                ):
        "Initialize Canvas Group within a Group Set and its appropriate memberships"
        self.API_URL = API_URL
        self.canvas = None
        self.course = None
        self.group_category = None
        self.group_categories = None
        self.groups = None
        self.users = None
        self.email_to_canvas_id = None
        self.canvas_id_to_email = None
        self.email_to_name = None
        self.API_KEY = None
        self.github = None
        self.credentials_fp = None
        self.groups = None
        self.group_to_emails = None
        self.assignment = None
        self.verbosity = verbosity
        
        # initialize by the input parameter
        if credentials_fp != "":
            self.auth_canvas(credentials_fp)
        if course_id != "":
            self.set_course(course_id)
            self.get_group_categories()
        if group_category != "":
            self.set_group_category(group_category)
            
    def auth_canvas(self,
                    credentials_fp: str # the Authenticator key generated from canvas
                   ):
        "Authorize the canvas module with API_KEY"
        self.credentials_fp = credentials_fp
        with open(credentials_fp, "r") as f:
            credentials = json.load(f)
        self.API_KEY = credentials["Canvas Token"]
        self.canvas = Canvas(self.API_URL, self.API_KEY)
        # test authorization
        _ = self.canvas.get_activity_stream_summary()
        if self.verbosity != 0:
            print(f"{bcolors.OKGREEN}Authorization Successful!{bcolors.ENDC}")
        
    def set_course(self, 
                   course_id: int # the course id of the target course
                  ):
        "Set the target course by the course ID"
        self.course = self.canvas.get_course(course_id)
        if self.verbosity != 0:
            print(f"Course Set: {bcolors.OKGREEN} {self.course.name} {bcolors.ENDC}")
            print(f"Getting List of Users... This might take a while...")
        self.users = list(self.course.get_users(enrollment_type=['student']))
        if self.verbosity != 0:
            print(f"Users Fetch Complete! The course has {bcolors.OKBLUE}{len(self.users)}{bcolors.ENDC} students.")
        self.email_to_canvas_id = {}
        self.canvas_id_to_email = {}
        self.email_to_name = {}
        for u in self.users:
            try:
                self.email_to_canvas_id[u.email.split("@")[0]] = u.id
                self.canvas_id_to_email[u.id] = u.email.split("@")[0]
                self.email_to_name[u.email.split("@")[0]] = u.short_name
            except Exception:
                if self.verbosity != 0:
                    print(f"{bcolors.WARNING}Failed to Parse email and id"
                          f" for {bcolors.UNDERLINE}{u.short_name}{bcolors.ENDC}{bcolors.ENDC}")

    def link_assignment(self,
                        assignment_id: int # assignment id, found at the url of assignmnet tab
                       ) -> canvasapi.assignment.Assignment: # target assignment
        "Link the target assignment on canvas"
        assignment = self.course.get_assignment(assignment_id)
        if self.verbosity != 0:
            print(f"Assignment {bcolors.OKGREEN+assignment.name+bcolors.ENDC} Link!")
        self.assignment = assignment
        return assignment

    def post_grade(self,
                    student_id: int, # canvas student id of student. found in self.email_to_canvas_id
                    grade: float, # grade of that assignment
                    text_comment="", # text comment of the submission. Can feed
                    force=False, # whether force to post grade for all students. If False (default), it will skip post for the same score.
                  ) -> canvasapi.submission.Submission: # created submission
        "Post grade and comment to canvas to the target assignment"
        submission = self.assignment.get_submission(student_id)
        if not force and submission.score == grade:
            if self.verbosity != 0:
                print(f"Grade for {bcolors.OKGREEN+self.canvas_id_to_email[student_id]+bcolors.ENDC} did not change.\n"
                      f"{bcolors.OKCYAN}Skipped{bcolors.ENDC}.\n"
                     )
            return
        edited = submission.edit(
            submission={
                'posted_grade': grade
            }, comment={
                'text_comment': text_comment
            }
        )
        if self.verbosity != 0:
            print(f"Grade for {bcolors.OKGREEN+self.canvas_id_to_email[student_id]+bcolors.ENDC} Posted!")
        return edited

    def get_email_by_name(self,
                          name_fussy: str # search by first name or last name of a student
                         ) -> str: # email of a search student
        name_fussy = name_fussy.lower()
        for email, name in self.email_to_name.items():
            if name_fussy in name.lower():
                return email
        raise ValueError(f"Name {name_fussy} Not Found.")
        
        
    def set_group_category(self,
                           category_name: str # the target group category
                          ) -> canvasapi.group.GroupCategory: # target group category object
        _ = self.get_group_categories()
        try:
            self.group_category = self.group_categories[category_name]
        except KeyError:
            raise KeyError(f"{category_name} did not found in the group categories. "
                           f"Try to create one with CanvasGroup.create_group_category")
        if self.verbosity != 0:
            print(f"Setting Group Category... ")
        self.groups = list(self.group_category.get_groups())
        self.group_to_emails = {
            group.name: [
                u.login_id for u in list(group.get_users())
            ] for group in self.groups}
        if self.verbosity != 0:
            print(f"Group Category: {bcolors.OKGREEN+category_name+bcolors.ENDC} Set!")
        return self.group_category
    
    def get_groups(self,
                   category_name="" # the target group category. If not provided, will look for self.group_category
                  ) -> dict: # {group_name: [student_emails]}
        if category_name != "":
            self.set_group_category(category_name)
            return self.group_to_emails
        if self.group_category is None:
            raise ValueError("Group Category is not set")
        return self.group_to_emails
        
    def get_course(self):
        return self.course
    
    def get_group_categories(self) -> dict: # return a name / group category object
        "Grab all existing group categories (group set) in this course"
        categories = list(self.course.get_group_categories())
        self.group_categories = {cat.name: cat for cat in categories}
        return {cat.name: cat for cat in categories}
    
    def create_group_category(self,
                              params: dict # the parameter of canvas group category API @ [this link](https://canvas.instructure.com/doc/api/group_categories.html#method.group_categories.create)
                             ) -> canvasapi.group.GroupCategory: # the generated group category object
        "Create group category (group set) in this course"
        self.group_category = self.course.create_group_category(**params)
        return self.group_category
    
    def create_group(self,
                     params: dict, #the parameter of canvas group create API at [this link](https://canvas.instructure.com/doc/api/groups.html#method.groups.create)
                    ) -> canvasapi.group.Group: # the generated target group object
        "Create canvas group under the target group category"
        if self.group_category is None:
            raise ValueError("Have you specified or create a group category (group set)?")
        group = self.group_category.create_group(**params)
        if self.verbosity != 0:
            print(f"In Group Set: {bcolors.OKBLUE+self.group_category.name+bcolors.ENDC},")
            print(f"Group {bcolors.OKGREEN+params['name']+bcolors.ENDC} Created!")
        return group
    
    def join_canvas_group(self,
                          group: canvasapi.group.Group, # the group that students will join
                          group_members:[str], # list of group member's SIS Login (email prefix, before the @.)
                         ) -> [str]: # list of unsuccessful join
        "Add membership access of each group member into the group"
        unsuccessful_join = []
        for group_member in group_members:
            try:
                canvas_id = self.email_to_canvas_id[group_member]
                group.create_membership(canvas_id)
                if self.verbosity != 0:
                    print(f"Member {bcolors.OKGREEN}{group_member}{bcolors.ENDC} Joined group {bcolors.OKGREEN}{group.name}{bcolors.ENDC}")
            except KeyError as e:
                unsuccessful_join.append(group_member)
                print(f"Error adding student {bcolors.WARNING+group_member+bcolors.ENDC} \n into group {group.name}")
                print(e)
        return unsuccessful_join
    
    def fetch_username_from_quiz(self,
                                 quiz_id: int, # quiz id of the username quiz
                                 col_index=7, # canvas quiz generated csv's question field column index
                                ) -> dict: # {SIS Login ID: github username} dictionary
        "Fetch the GitHub user name from the canvas quiz"
        header = {'Authorization': 'Bearer ' + self.API_KEY}
        quiz = self.course.get_quiz(quiz_id)
        if self.verbosity != 0:
            print(f"Quiz: {bcolors.OKGREEN+quiz.title+bcolors.ENDC} "
                  f"fetch! \nGenerating Student Analaysis..."
                 )
        report = quiz.create_report("student_analysis")
        progress_url = report.progress_url
        completed = False
        while not completed:
            status = requests.get(progress_url, headers = header).json()
            if self.verbosity != 0:
                self._progress(status["completion"])
                time.sleep(0.1)
            if status["completion"] == 100:
                completed = True
        if self.verbosity != 0:
            print(f"\n{bcolors.OKGREEN}Report Generated!{bcolors.ENDC}")
        # use requests to download the file 
        file_url = quiz.create_report("student_analysis").file["url"]
        response = requests.get(file_url, headers=header)
        file = StringIO(response.content.decode())
        # use pandas to parse the response csv
        df = pd.read_csv(file, delimiter=",")
        col = list(df.columns)
        # rename column
        if self.verbosity != 0:
            print(f"The Question asked is {bcolors.OKBLUE}{col[col_index]}{bcolors.ENDC}. \n"
                  f"Make sure this is the correct question where you asked student for their GitHub id.\n"
                  f"If you need to change the index of columns, change the col_index argument of this call."
                 )
        col[col_index] = "GitHub Username"
        df.columns = col
        small = df[["id", "GitHub Username"]].copy()
        small["email"] = small["id"].apply(lambda x: self.canvas_id_to_email[x])
        small = small[["email", "GitHub Username"]].set_index("email")
        return small.to_dict()["GitHub Username"]
    
    def _check_single_github_username(self,
                              email:str, # Student email
                              github_username:str, # student input we want to test
                             ) -> bool: # whether the username is valid
        "Check a single GitHub username on GitHub"
        if self.credentials_fp is None:
            raise ValueError("Credentials not set. Set it via self.auth_canvas")
        if self.github is None:
            # if the GitHub object has not been initialized
            with open(self.credentials_fp, "r") as f:
                credentials = json.load(f)
            github_token = credentials["GitHub Token"]
            self.github = Github(github_token)
        try:
            self.github.get_user(github_username)
        except Exception as e:
            print(f"User: {bcolors.WARNING+github_username+bcolors.ENDC} Not Found on GitHub")
            return False
        return True

    def check_github_usernames(self,
                               github_usernames:dict, # {email: github username} of student inputs, generated from self.fetch_username_from_quiz
                               send_canvas_email=False, # whether send a reminder for students who have an invalid GitHub username
                               send_undone_reminder=False, # send quiz undone reminder using canvas email
                               quiz_url="", # include a quiz url in the conversation for student to quickly complete the quiz.
                              ) -> dict: # {email: github username} of unreasonable GitHub id
        "batch check GitHub username from student inputs."
        unsuccessful = {}
        for email, github_username in github_usernames.items():
            valid = self._check_single_github_username(email, github_username)
            if not valid:
                unsuccessful[email] = github_username
                if send_canvas_email:
                    self.create_conversation(
                        self.email_to_canvas_id[email],
                        subject="Unidentifiable GitHub Username",
                        body=(f"Hi {email}, \n Your GitHub Username: {github_username} "
                              f"is unidentifiable on github.com. \n Please complete the quiz GitHub Username Quiz again.\n"
                              f"{quiz_url} \n"
                              f"Thank You."
                             )
                    )
                    if self.verbosity != 0:
                        print(f"{bcolors.OKGREEN}Notification Sent!{bcolors.ENDC}")
        if send_undone_reminder:
            submitted = github_usernames.keys()
            for email in self.email_to_canvas_id.keys():
                if email not in submitted:
                    if self.verbosity != 0:
                        print(f"Student {bcolors.WARNING}{email}{bcolors.ENDC} did not"
                              f" submit their github username."
                             )
                    # means student did not submit the quiz
                    if send_canvas_email:
                        self.create_conversation(
                            self.email_to_canvas_id[email],
                            subject="GitHub Username Quiz Not Completed",
                            body=(f"Hi {email}, \n You did not complete the GitHub Quiz."
                                  f"\n Please complete the quiz GitHub Username Quiz ASAP\n"
                                  f"{quiz_url} \n"
                                  f"Thank You."
                                 )
                        )
                        if self.verbosity != 0:
                            print(f"{bcolors.OKGREEN}Notification Sent!{bcolors.ENDC}")
        return unsuccessful
    
    def _progress(self,
                  percentage:int # percentage of the progress
                 ):
        sys.stdout.write('\r')
        # the exact output you're looking for:
        sys.stdout.write("[%-20s] %d%%" % ('='*int(percentage//5), percentage))
        sys.stdout.flush()
    
    def assign_canvas_group(self,
                            group_name: str, # group name, display on canvas
                            group_members:[str], # list of group member's SIS Login
                            in_group_category: str, # specify which group category the group belongs to
                           ) -> (canvasapi.group.Group, [str]): # list of unsuccessful join
        "Create new groups and assign group member into the class in the `self.group_category`"
        self.set_group_category(in_group_category)
        group = self.create_group({"name": group_name})
        unsuccessful_join = self.join_canvas_group(group, group_members)
        if self.verbosity != 0:
            print(f"Group {bcolors.OKGREEN+group_name+bcolors.ENDC} created!")
        return group, unsuccessful_join
    
    def create_conversation(self,
                            recipients:int, #  recipient ids. These may be user ids or course/group ids prefixed with ‘course_’ or ‘group_’ respectively.
                            subject:str, # subject of the conversation
                            body:str, # The message to be sent
                           ) -> canvasapi.conversation.Conversation: # created conversation
        "Create a conversation with the target user"
        conv = self.canvas.create_conversation(
            [recipients],
            body=body,
            subject=subject,
            context_code=f"course_{self.course.id}",
        )
        return conv




# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/CanvasGroupy',
                'doc_host': 'https://FleischerResearchLab.github.io',
                'git_url': 'https://github.com/FleischerResearchLab/CanvasGroupy',
                'lib_path': 'CanvasGroupy'},
  'syms': { 'CanvasGroupy.assign': { 'CanvasGroupy.assign.AssignGroup': ('api/groupeng_assign.html#assigngroup', 'CanvasGroupy/assign.py'),
                                     'CanvasGroupy.assign.AssignGroup.__init__': ( 'api/groupeng_assign.html#assigngroup.__init__',
                                                                                   'CanvasGroupy/assign.py'),
                                     'CanvasGroupy.assign.AssignGroup.assign_groups': ( 'api/groupeng_assign.html#assigngroup.assign_groups',
                                                                                        'CanvasGroupy/assign.py'),
                                     'CanvasGroupy.assign.AssignGroup.create_canvas_group': ( 'api/groupeng_assign.html#assigngroup.create_canvas_group',
                                                                                              'CanvasGroupy/assign.py'),
                                     'CanvasGroupy.assign.AssignGroup.create_github_group': ( 'api/groupeng_assign.html#assigngroup.create_github_group',
                                                                                              'CanvasGroupy/assign.py')},
            'CanvasGroupy.canvas': { 'CanvasGroupy.canvas.CanvasGroup': ('api/canvas.html#canvasgroup', 'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.__init__': ( 'api/canvas.html#canvasgroup.__init__',
                                                                                   'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup._check_single_github_username': ( 'api/canvas.html#canvasgroup._check_single_github_username',
                                                                                                        'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup._progress': ( 'api/canvas.html#canvasgroup._progress',
                                                                                    'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.assign_canvas_group': ( 'api/canvas.html#canvasgroup.assign_canvas_group',
                                                                                              'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.auth_canvas': ( 'api/canvas.html#canvasgroup.auth_canvas',
                                                                                      'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.check_github_usernames': ( 'api/canvas.html#canvasgroup.check_github_usernames',
                                                                                                 'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.create_conversation': ( 'api/canvas.html#canvasgroup.create_conversation',
                                                                                              'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.create_group': ( 'api/canvas.html#canvasgroup.create_group',
                                                                                       'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.create_group_category': ( 'api/canvas.html#canvasgroup.create_group_category',
                                                                                                'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.fetch_username_from_quiz': ( 'api/canvas.html#canvasgroup.fetch_username_from_quiz',
                                                                                                   'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.get_course': ( 'api/canvas.html#canvasgroup.get_course',
                                                                                     'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.get_email_by_name': ( 'api/canvas.html#canvasgroup.get_email_by_name',
                                                                                            'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.get_group_categories': ( 'api/canvas.html#canvasgroup.get_group_categories',
                                                                                               'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.get_groups': ( 'api/canvas.html#canvasgroup.get_groups',
                                                                                     'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.join_canvas_group': ( 'api/canvas.html#canvasgroup.join_canvas_group',
                                                                                            'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.link_assignment': ( 'api/canvas.html#canvasgroup.link_assignment',
                                                                                          'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.post_grade': ( 'api/canvas.html#canvasgroup.post_grade',
                                                                                     'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.set_course': ( 'api/canvas.html#canvasgroup.set_course',
                                                                                     'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.CanvasGroup.set_group_category': ( 'api/canvas.html#canvasgroup.set_group_category',
                                                                                             'CanvasGroupy/canvas.py'),
                                     'CanvasGroupy.canvas.bcolors': ('api/canvas.html#bcolors', 'CanvasGroupy/canvas.py')},
            'CanvasGroupy.github': { 'CanvasGroupy.github.GitHubGroup': ('api/github.html#githubgroup', 'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.__init__': ( 'api/github.html#githubgroup.__init__',
                                                                                   'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.add_collaborator': ( 'api/github.html#githubgroup.add_collaborator',
                                                                                           'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.add_team': ( 'api/github.html#githubgroup.add_team',
                                                                                   'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.auth_github': ( 'api/github.html#githubgroup.auth_github',
                                                                                      'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.create_feedback_dir': ( 'api/github.html#githubgroup.create_feedback_dir',
                                                                                              'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.create_group_repo': ( 'api/github.html#githubgroup.create_group_repo',
                                                                                            'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.create_issue': ( 'api/github.html#githubgroup.create_issue',
                                                                                       'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.create_issue_from_md': ( 'api/github.html#githubgroup.create_issue_from_md',
                                                                                               'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.create_repo': ( 'api/github.html#githubgroup.create_repo',
                                                                                      'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.get_org_repo': ( 'api/github.html#githubgroup.get_org_repo',
                                                                                       'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.get_repo': ( 'api/github.html#githubgroup.get_repo',
                                                                                   'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.get_team': ( 'api/github.html#githubgroup.get_team',
                                                                                   'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.release_feedback': ( 'api/github.html#githubgroup.release_feedback',
                                                                                           'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.remove_collaborator': ( 'api/github.html#githubgroup.remove_collaborator',
                                                                                              'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.rename_files': ( 'api/github.html#githubgroup.rename_files',
                                                                                       'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.resend_invitations': ( 'api/github.html#githubgroup.resend_invitations',
                                                                                             'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.resent_invitations_team_repos': ( 'api/github.html#githubgroup.resent_invitations_team_repos',
                                                                                                        'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.GitHubGroup.set_org': ( 'api/github.html#githubgroup.set_org',
                                                                                  'CanvasGroupy/github.py'),
                                     'CanvasGroupy.github.bcolors': ('api/github.html#bcolors', 'CanvasGroupy/github.py')},
            'CanvasGroupy.grading': { 'CanvasGroupy.grading.Grading': ('api/project_grading.html#grading', 'CanvasGroupy/grading.py'),
                                      'CanvasGroupy.grading.Grading.__init__': ( 'api/project_grading.html#grading.__init__',
                                                                                 'CanvasGroupy/grading.py'),
                                      'CanvasGroupy.grading.Grading.create_issue_from_md': ( 'api/project_grading.html#grading.create_issue_from_md',
                                                                                             'CanvasGroupy/grading.py'),
                                      'CanvasGroupy.grading.Grading.fetch_issue': ( 'api/project_grading.html#grading.fetch_issue',
                                                                                    'CanvasGroupy/grading.py'),
                                      'CanvasGroupy.grading.Grading.grade_project': ( 'api/project_grading.html#grading.grade_project',
                                                                                      'CanvasGroupy/grading.py'),
                                      'CanvasGroupy.grading.Grading.parse_score_from_issue': ( 'api/project_grading.html#grading.parse_score_from_issue',
                                                                                               'CanvasGroupy/grading.py'),
                                      'CanvasGroupy.grading.Grading.update_canvas_score': ( 'api/project_grading.html#grading.update_canvas_score',
                                                                                            'CanvasGroupy/grading.py'),
                                      'CanvasGroupy.grading.bcolors': ('api/project_grading.html#bcolors', 'CanvasGroupy/grading.py')}}}
